/***************************************************************************
                          mypoly3d.h  -  description
                             -------------------
    begin                : Thu Jun 28 2001
    copyright            : (C) 2001 by Rdiger Knrig
    email                : ruediger@knoerig.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MYPOLY3D_H
#define MYPOLY3D_H


#include <string.h>

/**
	* provides data structures for modelling
	* 3d objects bases on triangles
	* the 3D-system is a right-oriented (X and Y in screen orientaion, Z points in the screen)
  *	@author Rdiger Knrig
  */

class MyPoly3D {
public: 
	MyPoly3D() {}
	~MyPoly3D() {}
	/** structure representing a color in red,green,blue */
	struct color3f { float r,g,b; };
	/** structure representing a polygon */
	struct poly {int p1,p2,p3; int cindex; };
	/** class representing a point in homogeneous 3D-coordinates */
	class point4f;
	/** class representing the transformation matrix */
	class tmatrix;
};

static const float entity[4][4] = {{1.0,0.0,0.0,0.0},
											 			 			 {0.0,1.0,0.0,0.0},
											  		       {0.0,0.0,1.0,0.0},
											 			       {0.0,0.0,0.0,1.0}};
											 			
											 			
/**
	* This class provides a 4x4 transformation matrix for both view and object transformations
	* @author Rdiger Knrig
	*/
class MyPoly3D::tmatrix
{
	friend MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);
	
	public:
		/** supported view transformations */
		enum viewtransfrom_t {orthoXY,orthoYZ,orthoXZ,persp1};
		/** default constructor, generates an identity matrix */
		tmatrix() { reset(); }
		/** copy constructor */
		tmatrix(const MyPoly3D::tmatrix &src) { memcpy(matrix,src.matrix,16*sizeof(float)); }
		/** default destruktor */
		virtual ~tmatrix() {}
		/** resets matrix to entity matrix */
		void reset() { memcpy(matrix,entity,16*sizeof(float)); }
		/**
			* set view transformation type
			* @param type View transformation type
			* @param bw Optional focal length of camera lens (perspective transformation)
			*/
		void setTransform(viewtransfrom_t type, float bw = 1.0);
		
		/**
			* add a translation to the matrix
			* @param x X component of translation vector
			* @param y Y component of translation vector
			* @param z Z component of translation vector
			*/
		void translate(float x, float y, float z);
		
		/**
			* add a scaling operation to the matrix
			* @param x X scaling factor
			* @param y Y scaling factor
			* @param z Z scaling factor
			*/
		void scale(float,float y,float z);
		
		/**
			* Rotation around x axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundX(float degrees);
		
		/**
			* Rotation around y axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundY(float degrees);
		
		/**
			* Rotation around z axis
			* @param degrees Rotation angle in degrees
			*/
		void rotateAroundZ(float degrees);
		
		/** multiplication: matrix = matrix * matrix */
		MyPoly3D::tmatrix operator*(MyPoly3D::tmatrix b);
		
	protected:
		float matrix[4][4];
};

/** multiplication: vector = vector * matrix */
MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);


/**
	* this class represents a point in homogenous 3D coordinates
	* @author Rdiger Knrig
	*/
class MyPoly3D::point4f
{
	friend class MyPoly3D::tmatrix;
	friend MyPoly3D::point4f operator*(MyPoly3D::point4f v, MyPoly3D::tmatrix m);
	
	public:
		/** default constructor, results in a 0/0/0/0 - point */
		point4f() { memset(data,0,4*sizeof(float)); }
		/** copy constructor */
		point4f(const MyPoly3D::point4f &src) { memcpy(data,src.data,4*sizeof(float)); }
		/** constructor from 4 float values (w is optional ) */
		point4f(float x, float y, float z, float w=1.0) { data[0]=x; data[1]=y; data[2]=z; data[3]=w; }
		/** default destructor */
		virtual ~point4f() {}
		/** method for access to x component */
		float &x() {return data[0];}
		/** method for access to y component */
		float &y() {return data[1];}
		/** method for access to z component */
		float &z() {return data[2];}
		/** method for access to w component */
		float &w() {return data[3];}
		/** substraction this - b = result */
		point4f operator-(point4f b) {return point4f(data[0]/data[3]-b.data[0]/b.data[3],data[1]/data[3]-b.data[1]/b.data[3],data[2]/data[3]-b.data[2]/b.data[3],1.0); }
		/** addition this + b = result */
		point4f operator+(point4f b) {return point4f(data[0]/data[3]+b.data[0]/b.data[3],data[1]/data[3]+b.data[1]/b.data[3],data[2]/data[3]+b.data[2]/b.data[3],1.0); }		
	protected:
		float data[4];	
};


#endif
