/***************************************************************************
                          qtplgviewerview.cpp  -  description
                             -------------------
    begin                : Don Jun 28 10:55:59 CEST 2001
    copyright            : (C) 2001 by Rdiger Knrig
    email                : ruediger@knoerig.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qprinter.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qrect.h>
#include <qpoint.h>
#include <qpen.h>
#include <qbrush.h>
#include <qmessagebox.h>
#include <qapplication.h>

// application specific includes
#include "qtplgviewerview.h"
#include "qtplgviewerdoc.h"



QtpLgviewerView::QtpLgviewerView(QtpLgviewerDoc* pDoc, QWidget *parent, const char* name, int wflags)
 : QWidget(parent, name, wflags)
{
		setFocusPolicy(QWidget::StrongFocus);
		
		connect(this,SIGNAL(signalViewSettings()),\
		qApp->mainWidget(),SLOT(slotUpdateViewToolbars()));
		
		mycaption = caption();
		
    doc=pDoc;
    viewtransform = MyPoly3D::tmatrix::orthoXY;
    viewtransf.setTransform(MyPoly3D::tmatrix::orthoXY);
    viewtype = wireframe;
		bw = 2.0;
		scale = 8.0;
		slotUpdateColorTable();
		backgroundcolor.setRgb(0,0,0);
		setBackgroundColor(backgroundcolor);
		backbuffer.resize(width(),height());
		backbuffer.fill(backgroundcolor);
		projpolygon = QPointArray(3);
		repaint();
}

QtpLgviewerView::~QtpLgviewerView()
{
}

QtpLgviewerDoc *QtpLgviewerView::getDocument() const
{
	return doc;
}

void QtpLgviewerView::update(QtpLgviewerView* pSender){
	if(pSender != this)
	{
		slotUpdateColorTable();
		slotUpdateBackBuffer();
		repaint();
	}
}

void QtpLgviewerView::print(QPrinter *pPrinter)
{
  if (pPrinter->setup(this))
  {
		QPainter p;
		p.begin(pPrinter);
		while(!backbuffermutex.lock());
		p.drawPixmap(0,0,backbuffer);
		backbuffermutex.unlock();
		p.end();
  }
}

void QtpLgviewerView::closeEvent(QCloseEvent*)
{
  // LEAVE THIS EMPTY: THE EVENT FILTER IN THE QtpLgviewerApp CLASS TAKES CARE FOR CLOSING
  // QWidget closeEvent must be prevented.
}

/** rereads color table from document */
void QtpLgviewerView::slotUpdateColorTable()
{
		bool colorproblem = false;
	  //lock access to document data and build color table
		while(!doc->accessmutex.lock());
		
		colortable.resize(doc->colortable.size());
		vector<MyPoly3D::color3f>::iterator src = doc->colortable.begin();
		for(vector<QColor>::iterator dst = colortable.begin(); dst != colortable.end(); dst++,src++)
		{
			dst->setRgb(int(255*src->r),int(255*src->g),int(255*src->b));
			if(*dst == backgroundcolor) colorproblem = true;
		}
		// unlock access to documents data
		doc->accessmutex.unlock();
		if(colorproblem) QMessageBox::information(this,"warning:","An object color is identical to the background color\n"
		"Suggest switching to another background color");
}	

void QtpLgviewerView::paintEvent(QPaintEvent *p)
{
	if(backbuffer.isNull()) return;
	
	QRect viewportarea(0,0,width(),height());
	/** union between viewport rect and paintevent area should be updated */
	viewportarea.intersect(p->rect());
	if(viewportarea.isNull()) return;
	
	while(!backbuffermutex.lock());
	bitBlt(this,viewportarea.topLeft(),&backbuffer,viewportarea,Qt::CopyROP);
	backbuffermutex.unlock();
}

void QtpLgviewerView::resizeEvent(QResizeEvent *r)
{
		while(!backbuffermutex.lock());
		backbuffer.resize(r->size());	
		backbuffermutex.unlock();
		slotUpdateBackBuffer();
}

/** repaints the backbuffer */
void QtpLgviewerView::slotUpdateBackBuffer()
{
	int xoffset = backbuffer.width() >> 1;
	int yoffset = backbuffer.height() >> 1;
	
	//lock access to document data
	while(!doc->accessmutex.lock());
	// lock access to backbuffer
	while(!backbuffermutex.lock());
	
	backbuffer.fill(backgroundcolor);
	QPainter p(&backbuffer);
	p.setClipRect(0,0,backbuffer.width(),backbuffer.height());
	
			
	// iterate through all polygons
	for(vector<MyPoly3D::poly>::iterator poly = doc->polylist.begin(); poly != doc->polylist.end(); poly++)
	{
			polygon[0] = doc->pointlist[poly->p1];
			polygon[1] = doc->pointlist[poly->p2];
			polygon[2] = doc->pointlist[poly->p3];
			
			// transformate points in the camera oriented coordinate system
			for(int i=0; i < 3; i++) polygon[i] = polygon[i] * objtransf;
			
			// do not draw polygons where a point has a z coordinate <= -bw
			if(viewtransform == MyPoly3D::tmatrix::persp1)
			{
				if(polygon[0].z() <= -bw || polygon[1].z() <= -bw || polygon[2].z() <= -bw) continue;
				for(int i=0; i < 3; i++)
				{
					polygon[i].x() /= polygon[i].w();
					polygon[i].y() /= polygon[i].w();
					polygon[i].z() /= polygon[i].w();
				}
			}
			
			if(viewtype & backfaceculling)
			{
				MyPoly3D::point4f edge1 = polygon[1] - polygon[0];
				MyPoly3D::point4f edge2 = polygon[2] - polygon[0];
				if(edge1.x()*edge2.y() - edge1.y()*edge2.x() >= 0.0) continue; // norm vector of triangle heads "away"
			}
			
			// apply view transformation
			for(int i=0; i < 3; i++) polygon[i] = polygon[i] * viewtransf;
			
			// add points to QPointArray
			for(int i=0; i < 3; i++) projpolygon.setPoint(i,xoffset + int(scale*polygon[i].x()/polygon[i].w()),yoffset - int(scale * polygon[i].y()/polygon[i].w()));
			
			// set wireframe color
			if(viewtype & wireframe)
			{
				if(viewtype & flatshading)
				{
					// use the opposite color (rgb system) to the background color
					p.setPen(QPen(QColor(255-backgroundcolor.red(),255-backgroundcolor.green(),255-backgroundcolor.blue())));
				}
				else
				{
					// use polygon color
					p.setPen(QPen(colortable[poly->cindex]));
				}
			}
			else
			{
				p.setPen(QPen(Qt::NoPen));
			}		
			
			// set face color
			if(viewtype & flatshading)
			{
				p.setBrush(QBrush(colortable[poly->cindex]));
			}
			else
			{
				p.setBrush(QBrush(QBrush::NoBrush));
			}
			
			// draw the polygon
			p.drawPolygon(projpolygon);
	}
	
	doc->accessmutex.unlock();
	backbuffermutex.unlock();
}

void QtpLgviewerView::slotSetViewTransform(MyPoly3D::tmatrix::viewtransfrom_t viewtransform)
{
	while(!backbuffermutex.lock());
	this->viewtransform = viewtransform;
	viewtransf.setTransform(viewtransform,bw);
	backbuffermutex.unlock();
	slotUpdateBackBuffer();
	repaint();
	slotUpdateName();
}

void QtpLgviewerView:: keyPressEvent(QKeyEvent *k)
{
	Qt::ButtonState modifier = k->state();
	switch(k->key())
	{
		case Key_Plus : slotScaleUp();
			break;
		case Key_Minus : slotScaleDown();
			break;
		case Key_Enter : slotScaleOnetoOne();
			break;
		case Key_Up :
			switch(modifier)
			{
				case ShiftButton : slotLookUp();
					break;
				case AltButton :  slotStepUp();
					break;
				default: slotStepForward();
			}
			break;
		case Key_Down :
			switch(modifier)
			{
				case ShiftButton : slotLookDown();
					break;
				case AltButton :  slotStepDown();
					break;
				default: slotStepBackward();
			}
			break;
		case Key_Left :
			switch(modifier)
			{
				case ShiftButton : slotLookLeft();
					break;
				case AltButton :
				default:					slotStepLeft();
			}
			break;
		case Key_Right :
			switch(modifier)
			{
				case ShiftButton : slotLookRight();
					break;
				case AltButton :
				default:					slotStepRight();
			}
		 break;
		default: break;
	}
}

/** No descriptions */
void QtpLgviewerView::slotUpdateName(const QString &basename)
{
	static const QString viewtransfrom_names[4] = {" orthoXY "," orthoYZ "," orthoXZ "," persp1 "};
	
	if(!basename.isEmpty()) mycaption = basename;
	setCaption(mycaption + viewtransfrom_names[(int) viewtransform]);
}
